@@ -70,9 +70,13 @@
 /* for strcasestr */
 #define _GNU_SOURCE
 
+#include <ctype.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 
+#include <regex.h>
+
 #include <X11/Xlib.h>
 #include <X11/extensions/Xcomposite.h>
 #include <GL/gl.h>
@@ -86,6 +90,54 @@ _print_error (const char *str)
         fprintf (stderr, "gnome-session-is-accelerated: %s\n", str);
 }
 
+static int
+_parse_kcmdline (void)
+{
+        FILE *kcmdline;
+        char *line = NULL;
+        size_t line_len = 0;
+        int ret = -1;
+
+        kcmdline = fopen("/proc/cmdline", "r");
+        if (kcmdline == NULL)
+                return ret;
+
+        while (getline (&line, &line_len, kcmdline) != -1) {
+                const char *arg;
+                const char *str;
+                int key_len = strlen ("gnome.fallback=");
+
+                if (line == NULL)
+                        break;
+
+                /* don't break if we found the argument once: last mention wins */
+
+                str = line;
+                do {
+                        arg = strstr (str, "gnome.fallback=");
+                        str = arg + key_len;
+
+                        if (arg &&
+                                        (arg == line || isspace (arg[-1])) && /* gnome.fallback= is really the beginning of an argument */
+                                        (isdigit (arg[key_len]))) { /* the first character of the value of this argument is an integer */
+                                if ((arg[key_len+1] == '\0' || isspace (arg[key_len+1]))) /* the value of this argument is only one character long */
+                                        ret = arg[key_len] - '0';
+                                else /* invalid value */
+                                        ret = 0xDEAD;
+
+                        }
+                } while (arg != NULL);
+
+                free (line);
+                line = NULL;
+                line_len = 0;
+        }
+
+        fclose (kcmdline);
+
+        return ret;
+}
+
 static int
 _has_composite (Display *display)
 {
@@ -97,6 +149,92 @@ _has_composite (Display *display)
         return 1;
 }
 
+static int
+_is_comment (const char *line)
+{
+        while (*line && isspace(*line))
+                line++;
+
+        if (*line == '#' || *line == '\0')
+                return 0;
+        else
+                return 1;
+}
+
+static int
+_is_gl_renderer_blacklisted (const char *renderer)
+{
+        FILE *blacklist;
+        char *line = NULL;
+        size_t line_len = 0;
+        int ret = 1;
+
+        blacklist = fopen(PKGDATADIR "/hardware-compatibility", "r");
+        if (blacklist == NULL)
+                goto out;
+
+        while (getline (&line, &line_len, blacklist) != -1) {
+                int whitelist = 0;
+                const char *re_str;
+                regex_t re;
+                int status;
+
+                if (line == NULL)
+                        break;
+
+                /* Drop trailing \n */
+                line[strlen(line) - 1] = '\0';
+
+                if (_is_comment (line) == 0) {
+                        free (line);
+                        line = NULL;
+                        continue;
+                }
+
+                if (line[0] == '+')
+                        whitelist = 1;
+                else if (line[0] == '-')
+                        whitelist = 0;
+                else {
+                        _print_error ("Invalid syntax in this line for hardware compatibility:");
+                        _print_error (line);
+                        free (line);
+                        line = NULL;
+                        continue;
+                }
+
+                re_str = line + 1;
+
+                if (regcomp (&re, re_str, REG_EXTENDED|REG_ICASE|REG_NOSUB) != 0) {
+                        _print_error ("Cannot use this regular expression for hardware compatibility:");
+                        _print_error (re_str);
+                } else {
+                        status = regexec (&re, renderer, 0, NULL, 0);
+                        regfree(&re);
+
+                        if (status == 0) {
+                                if (whitelist)
+                                        ret = 0;
+                                goto out;
+                        }
+                }
+
+                free (line);
+                line = NULL;
+        }
+
+        ret = 0;
+
+out:
+        if (line != NULL)
+                free (line);
+
+        if (blacklist != NULL)
+                fclose (blacklist);
+
+        return ret;
+}
+
 static int
 _has_hardware_gl (Display *display)
 {
@@ -143,12 +281,7 @@ _has_hardware_gl (Display *display)
                 goto out;
 
         renderer = (const char *) glGetString (GL_RENDERER);
-        /* The current Mesa software GL renderer string is
-	 * "Software Rasterizer".
-	 * Gallium has softpipe and llvmpipe. */
-        if (strcasestr (renderer, "software rasterizer") != NULL ||
-            strcasestr (renderer, "softpipe") != NULL ||
-            strcasestr (renderer, "llvmpipe") != NULL)
+        if (_is_gl_renderer_blacklisted (renderer) != 0)
                 goto out;
 
         /* we need to get the max texture size while we have a context,
@@ -257,9 +390,23 @@ _is_max_texture_size_big_enough (Display *display)
 int
 main (int argc, char **argv)
 {
+        int      kcmdline_parsed;
         Display *display = NULL;
         int      ret = 1;
 
+        kcmdline_parsed = _parse_kcmdline ();
+        if (kcmdline_parsed >= 0) {
+                if (kcmdline_parsed == 0) {
+                        _print_error ("Non-fallback mode forced by kernel command line.");
+                        ret = 0;
+                        goto out;
+                } else if (kcmdline_parsed == 1) {
+                        _print_error ("Fallback mode forced by kernel command line.");
+                        goto out;
+                } else
+                        _print_error ("Invalid value for gnome.fallback passed in kernel command line.");
+        }
+
         display = XOpenDisplay (NULL);
         if (!display) {
                 _print_error ("No X display.");
